package cz.muni.fi.xuhlarik.grammarManager.service;

import cz.muni.fi.xuhlarik.grammarManager.FileType;
import edu.cmu.sphinx.jsgf.JSGFGrammarException;
import edu.cmu.sphinx.jsgf.JSGFGrammarParseException;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammar;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarFactory;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarManager;
import edu.cmu.sphinx.jsgf.parser.JSGFParser;
import edu.cmu.sphinx.jsgf.parser.TokenMgrError;
import edu.cmu.sphinx.jsgf.rule.JSGFRule;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleName;
import org.jvoicexml.processor.grammar.Rule;
import org.jvoicexml.processor.srgs.GrammarException;
import org.jvoicexml.processor.srgs.xml.SrgsRuleGrammarParser;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {



    public  String getParsedFile(String file, FileType fileType) {
        if (fileType == null) {
            fileType = getFileExtension(file);

        }
        return parseFile(file, fileType);
    }


    //function validates xml document and transforms jsgf document to xml
    private  String parseFile(String file, FileType fileType) {

        if (fileType==FileType.XML_SRGS) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();
                InputSource is = new InputSource(new StringReader(file.trim()));
                saxParser.parse(is, new DefaultHandler());
                return file;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "xml parse error\n" + e.getMessage(), e);
            }

        }
        try {
            return jsgfToXml(file);
        } catch (JSGFGrammarParseException | TokenMgrError | IOException | URISyntaxException |
                 JSGFGrammarException e) {
            if (e instanceof JSGFGrammarParseException jsgfGrammarParseException) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + jsgfGrammarParseException.details);
            }
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + e.getMessage(), e);
        }


    }


    private  FileType getFileExtension(String file) {

        Pattern jsgf = Pattern.compile("^\\W*#JSGF.*;");
        Matcher mJsgf = jsgf.matcher(file);
        if (mJsgf.find()) {
            return FileType.JSGF;
        }
        Pattern xml = Pattern.compile("^\\W*<\\?xml.*\\?>");
        Matcher mXml = xml.matcher(file);
        if (mXml.find()) {
            return FileType.XML_SRGS;
        }

        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "format cant be found from file");
    }

    public  String xmlToJSGF(String xml, boolean grammar) {

        //String comments xd
        //todo language missing
        StringBuilder xmlString = new StringBuilder();
        SrgsRuleGrammarParser parser = new SrgsRuleGrammarParser();
        writeHeaderXml(xmlString, null, grammar);
        xmlString.append(xml);
        if (!grammar) {
            xmlString.append("</grammar>");
        }

        List<Rule> rules;
        try {
            rules = parser.load(new StringReader(xmlString.toString()));

        } catch (URISyntaxException | GrammarException e) {
            throw new ResponseStatusException(HttpStatus.INTERNAL_SERVER_ERROR, "error in converting xml to jsgf");
        }

        StringBuilder jsgfString = new StringBuilder();
        if (grammar) {
            Map<String, Object> parserAtributes = parser.getAttributes();
            writeHeaderJSGF(jsgfString, (String) parserAtributes.get("name"), (String) parserAtributes.get("xml:lang"));
            Object imports = parser.getAttributes().get("imports");
            if (imports != null) {
                jsgfString.append(imports);
            }
        }

        rules.forEach(p -> jsgfString.append(p.toStringJSGF()).append("\n"));
        return jsgfString.toString();
    }

    public String jsgfToXml(String file) throws JSGFGrammarParseException, IOException, URISyntaxException, JSGFGrammarException {
        JSGFRuleGrammar grammarFromJSGF = JSGFParser.newGrammarFromJSGF(new StringReader(file), new JSGFRuleGrammarFactory(new JSGFRuleGrammarManager()));
        StringBuilder sb = new StringBuilder();
        //todo
        writeHeaderXml(sb, null, false);
        sb.append("<meta name=\"Source\" content=\"").append(grammarFromJSGF.getName()).append(".jsgf\"/>\n");
        for (JSGFRuleName importName : grammarFromJSGF.getImports()) {
            sb.append("<meta name=\"seeAlso\" content=\"").append(importName.getRuleName()).append("\"/>\n");
        }


        for (String ruleName : grammarFromJSGF.getRuleNames()) {
            JSGFRule rule = grammarFromJSGF.getRule(ruleName);
            boolean publicRule = grammarFromJSGF.isRulePublic(ruleName);
            writeXmlRule(sb, publicRule, ruleName, rule.toStringXml());
        }
        sb.append("</grammar>");

        return sb.toString();
    }

    private void writeXmlRule(StringBuilder sb, boolean publicRule, String name, String rightSide) {
        sb.append("<rule ");
        sb.append("id=\"");sb.append(name);sb.append("\"");
        if (publicRule) {
            sb.append(" scope=\"public\"");
        }
        sb.append(">\n");
        sb.append(rightSide);
        sb.append("\n</rule>\n");
    }

    private  void writeHeaderXml(StringBuilder sb, String language, boolean justPrefix) {

        if (language == null) {
            language = "cz";
        }
        //xml
        sb.append("<?xml version=\"1.0\"?>\n");
        if (!justPrefix) {
            sb.append("<grammar version=\"1.0\"\n");
            sb.append("xml:lang='").append(language).append("'\n");
            sb.append("xmlns=\"http://www.w3.org/2001/06/grammar\">\n");
        }
    }

    private  void writeHeaderJSGF(StringBuilder sb, String name, String language) {
        sb.append("#JSGF V1.0").append(language == null ? "" : " " + language).append(";\n");
        sb.append("grammar ").append(name == null ? "defaulf" : name).append(";\n\n");
    }

}

